#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <string.h>
#include "id3lib.h"
#include "id3types.h"

using namespace std;

ID3Lib::ID3Lib () { _loaded = false; _version = 0; }

ID3Lib::ID3Lib (char *fileName)
{

	// Initialize our properties
	_loaded = false;
	_version = 0;

	// Open the file
	ifstream file(fileName, ios::in | ios::binary);
	if (file.fail()) {
		cout << "<message>File failed to open:" << errno << "</message>";
		return;
	}

	if (file.bad ()) {
		cout << "<message>Bad file mode</message>";
		return;
	}

	// Read the tag header
	file.read(reinterpret_cast<char *>(&fileHead), sizeof (fileHead));

	// Check for the ID3 tag head
	if (!strcmp (fileHead.ID3, "ID3")) {
		file.close ();
		return;
	}

	// No flag options supported at this time
	if ((int)fileHead.flags != 0) {
		file.close ();
		return;
	}

	// Get the correct header size
	unsigned long int size = calcSize (LSB(fileHead.size));
	unsigned long int bytesRead = 0;
	
	// Update the properties
	_version = fileHead.version;
	_loaded = true;

	// Load in the frames
	ID3FrameHeader frame;
	unsigned long int fSize, numFrames = 0;
	char *data;
	while (bytesRead < size) {
		
		// Grab and format the header information
		file.read(reinterpret_cast<char *>(&frame), sizeof (frame));
		fSize = LSB (frame.size); frame.size = LSB (frame.size);
		
		// If we came back with a size of zero, there's probably nothing else. Jump out
		if (fSize == 0)
			break;

		// Read in the data
		data = new char[fSize];
		file.read(data, fSize);

		// Add the frame to our frames vector
		frames.push_back (ID3Frame(&frame, data, fSize));
		
		// Iterate and repeat
		bytesRead += fSize + 10;

	}

	// Attempt to calculate the length of the song
	unsigned long int mp3head;
	file.seekg (size + 10);
	file.read (reinterpret_cast<char *>(&mp3head), sizeof (mp3head));
	
	// Validate the MP3 header
	mp3head = LSB (mp3head);
	if (((mp3head & 0xffe00000) >> 21) != 0x7ff) {
		delete [] data;
		file.close ();
		_loaded = false;
		return;
	}

	// Get the audio properties
	unsigned int mpegVersion = ((mp3head & 0x180000) >> 19);
	unsigned int mpegLayer = ((mp3head & 0x60000) >> 17);
	unsigned int bitrate = bitrates[((mp3head & 0xf000) >> 12)];
	unsigned int samples = ((mp3head & 0xc00) >> 10);
	unsigned int channels = ((mp3head & 0xc0) >> 6);

	// Check to see if there is an ID3v1 tag
	char id3v1[3];
	file.seekg (0, ios::end);
	unsigned long int len = file.tellg();
	file.seekg (len - 128);
	file.read (id3v1, sizeof (id3v1));
	
	// Compute the length of the MP3 audio data
	if (strcmp (id3v1, "TAG"))
		len -= (size + 10 + 128);
	else
		len -= (size + 10);

	// Calculate the length of the song
	_length = len / (bitrate / 8);

	// Clean up
	delete [] data;
	file.close ();

}

ID3Lib::~ID3Lib () { frames.clear (); }

string ID3Lib::tag (char *id)
{

	// If the provided ID is more than four characters long, save some cycles and jump out now
	if (strlen (id) > 4)
		return "\0";
	vector<ID3Frame>::iterator i;
	for (i = frames.begin(); i != frames.end(); ++i) {
		if (i->compare (id))
			return i->toString();
	}

	return "\0";

}

bool ID3Lib::saveAlbumArt (char *fileName)
{
	// Search for the APIC tag and save it if found
	vector<ID3Frame>::iterator i;
	for (i = frames.begin(); i != frames.end(); ++i) {
		if (i->compare ("APIC")) {
			i->save (fileName);
			return true;
		}
	}

	return false;
}

unsigned int ID3Lib::LSB (unsigned short int num) { return (num >> 8) | (num << 8); }
unsigned int ID3Lib::LSB (unsigned long int num) { return ((num & 0xff) << 24) | ((num & 0xff00) << 8) | ((num & 0xff0000) >> 8) | ((num & 0xff000000) >> 24); }
unsigned long int ID3Lib::calcSize (unsigned long int num) { return (num & 0x7f) | ((num & 0x7f00) >> 1) | ((num & 0x7f0000) >> 2) | ((num & 0x7f000000) >> 3); }