﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

namespace Id3Lib_Net
{

    public class Frame
    {

        private String _name;
        private byte[] _data;

        public Frame() { }

        public Frame(String name, byte[] data)
        {
            _name = name;
            _data = data;
        }

        public String name () { return _name; }
        public String getString () { return System.Text.ASCIIEncoding.ASCII.GetString(_data); }
        public byte[] getData() { return _data; }

    }

    public class Reader
    {

        // Properties
        public bool _loaded;
        private List<Frame> _frames = new List<Frame>();
        private String _title, _artist, _album, _year, _disc, _errstr;

        // Structs
        [StructLayout(LayoutKind.Explicit, Size=7)]
        private struct ID3FileHeader
        {
            [FieldOffset(0)]
            public UInt16 version;
            [FieldOffset(2)]
            public Byte flags;
            [FieldOffset(3)]
            public UInt32 size;
        }

        public Reader() { }

        /// <summary>
        /// Loads an MP3 file and parses the ID3v2 information
        /// </summary>
        /// <param name="fileName">Path to the MP3 file</param>
        public Reader(string fileName)
        {

            // Open the file
            FileStream stream;
            BinaryReader binRead;
            try {
                stream = new FileStream(fileName, FileMode.Open);
                binRead = new BinaryReader(stream);
            }
            catch {
                _loaded = false;
                _errstr = "Unable to open file: " + fileName;
                return;
            }

            // Various crap we need for reading the files
            byte[] readBuffer;
            Char[] ID3 = new Char[3], _frameID = new Char[4];
            String frameID;
            ID3FileHeader head;
            UInt32 frameSize, bytesRead = 0;
            UInt16 frameFlags;

            // Read in the identifier
            ID3 = binRead.ReadChars(3);

            // Validate the identifier
            if (ID3[0] != 'I' || ID3[1] != 'D' || ID3[2] != '3')
            {
                _errstr = "No ID3v2 tag found";
                _loaded = false;
                return;
            }

            // Read in the header
            int size = Marshal.SizeOf(typeof(ID3FileHeader));
            readBuffer = new byte[size];
            readBuffer = binRead.ReadBytes(size);
            GCHandle handle = GCHandle.Alloc(readBuffer, GCHandleType.Pinned);
            head = (ID3FileHeader)Marshal.PtrToStructure(handle.AddrOfPinnedObject(), typeof(ID3FileHeader));
            handle.Free();

            // Calculate the ID3 size
            UInt32 Id3Size = calcSize(head.size);

            // Read in all the frames
            while (bytesRead < Id3Size)
            {
                _frameID = binRead.ReadChars(4);
                frameID = new String(_frameID);
                frameSize = LSB (binRead.ReadUInt32());
                if (frameSize == 0)
                    break;
                frameFlags = binRead.ReadUInt16();
                readBuffer = new byte[frameSize];
                readBuffer = binRead.ReadBytes((int)frameSize);
                bytesRead += frameSize + 10;
                
                // Create a new frame and add it to the list
                _frames.Add(new Frame(frameID, readBuffer));

            }

            // Close the file stream
            binRead.Close();
            
            // Populate the common tags
            _loaded = true;
            populateCommon();

        }

        private void populateCommon ()
        {
            Frame f = getFrame("TIT2");
            if (f != null) _title = f.getString();

            f = getFrame("TALB");
            if (f != null) _album = f.getString();

            f = getFrame("TPE1");
            if (f != null) _artist = f.getString();

            f = getFrame("TPOS");
            if (f != null) _disc = f.getString();

            f = getFrame("TYER");
            if (f != null) _year = f.getString();
        }

        public String title() { return _title; }
        public String album() { return _album; }
        public String artist() { return _artist; }
        public String year() { return _year; }
        public String disc() { return _disc; }
        public String error() { return _errstr; }
        public bool fail() { return !_loaded; }

        /// <summary>
        /// Returns a frame based on the provided tag ID
        /// </summary>
        /// <param name="tagName">ID of the tag to search for</param>
        /// <returns></returns>
        public Frame getFrame(String tagName)
        {
            foreach (Frame frame in _frames)
            {
                if (frame.name() == tagName)
                    return frame;
            }
            return null;
        }

        /// <summary>
        /// Saves the APIC tag to the file sepcified (only works for JPEG data)
        /// </summary>
        /// <param name="file">Path to the file to save</param>
        /// <returns></returns>
        public bool savePicture(String file)
        {
            
            // Get the APIC frame
            Frame apic = getFrame("APIC");
            if (apic == null)
            {
                return false;
                _errstr = "No picture in MP3 file";
            }
            
            // Open the output file and write the contents
            try
            {
                Stream stream = new FileStream(file, FileMode.Create);
                BinaryWriter binOut = new BinaryWriter(stream);
                byte[] data = apic.getData();
                byte[] newData = new byte[1];
                bool found = false;

                // Search through the array for the beginning of a JPEG file
                if (data.Length >= 2)
                {
                    for (int i = 0; i < data.Length - 2; i++)
                    {
                        if (data[i] == 0xff && data[i + 1] == 0xd8)
                        {
                            newData = new byte[data.Length - i];
                            Buffer.BlockCopy(data, i, newData, 0, data.Length - i);
                            found = true;
                            break;
                        }
                    }
                    if (found)
                    {
                        binOut.Write(newData);
                        binOut.Close();
                        return true;
                    }
                    else
                    {
                        _errstr = "No JPEG data found in the MP3 file";
                        return false;
                    }
                }
                else
                {
                    _errstr = "No JPEG data found in the MP3 file";
                    return false;
                }
            }
            catch
            {
                _errstr = "There was an error writing the file";
                return false;
            }

        }

        private UInt32 calcSize(UInt32 num)
        {
            num = LSB(num);
            return (num & 0x7f) | ((num & 0x7f00) >> 1) | ((num & 0x7f0000) >> 2) | ((num & 0x7f000000) >> 3);
        }

        private UInt32 LSB(UInt16 num)
        {
            return (UInt32) ((num >> 8) | (num << 8));
        }

        private UInt32 LSB(UInt32 num)
        {
            return ((num & 0xff) << 24) | ((num & 0xff00) << 8) | ((num & 0xff0000) >> 8) | ((num & 0xff000000) >> 24);
        }

    }
}
